// LCD.c, routines to drive common LCD display off port E
//@@@ add watchdog clear to all this later if using watchdog timer

//MultiGeiger version
// main changes so far are use of CCP2 as input on porte.7

/////////////////////////////////////////////////////////
// doing things mostly "the CCS way" here
// avoiding the need for function prototypes by putting them in use order

#use fast_io (E) // we will control the vertical, the horizontal

#use fast_io (F) // for a second set of lines

#define PORTE 0xF84
#define TRISE 0xF96
#define PORTF 0xF85


#define LCD_CMD FALSE // which are we writing?
#define LCD_DATA TRUE // state of RS pin for data

#define LCD_PW 10 // "E" pulse width in CPU instructions
#define LCD_STL 5 // settling time in "" ""

#byte LCD_PORT = PORTE 
#bit LCD_EN = PORTE.6  //
//#bit LCD_EN = PORTF.7   // for second set of lines on a 4 liner
#bit LCD_RW = PORTE.5  //   
#bit  LCD_RS = PORTE.4 //  
#bit LCD_BSY = PORTE.3 // busy bit

/////////////////////////////////////////////////////////
void LCD_waitbusy(void)
{
 int1 tmp; // bit variable
 
 restart_WDT(); // just in case someone enabled it
 set_tris_e(0x8f); // @@@ might mess up E7 (let's always use it as output!)
 LCD_EN = FALSE; // which is a low-true signal...
 LCD_RS = FALSE;
 LCD_RW = TRUE; // this is a read operation, about the only one
 delay_us(100); // wait a bit for everything to turn around etc
               // large display with different controller doesn't
               // declare busy right away!!!  This can be removed
               // for the faster responding display controllers
 
 do {
  LCD_EN = TRUE; // force E high, to start cycle
  delay_cycles(LCD_PW + 2);// wait for slow lcd to see it
  tmp = LCD_BSY;  // get bit that indicates busy
  LCD_EN = FALSE;  // drop strobe, we got our bit
  delay_cycles(LCD_PW); // wait a bit
  LCD_EN = TRUE;   // dummy read of other nybble
  delay_cycles(LCD_PW); // 800ns delay, about right for slow LCD
  LCD_EN = FALSE;  //
  delay_cycles(LCD_PW); // a little delay before restarting loop
 } while (tmp != 0); // until ready. @@@ add timeout later
}
/////////////////////////////////////
void LCD_write(int8 data,int8 rs) // rs FALSE for command
{
 restart_WDT(); // just in case someone enabled it
 LCD_waitbusy(); // make sure we have a ready unit
 LCD_RS = rs; // both command and data, 0 for command, bit0=1 data
 LCD_RW = FALSE; // write to LCD (turn its bus to input)
 LCD_EN = FALSE; // make sure strobe is false
 delay_cycles (LCD_STL); // a small wait for robustness -- time to reverse drivers
 LCD_PORT &= 0xf0; // clear low part so we can or in the cmd/data bits
 set_tris_e(get_tris_e() & 0xf0); // make low 4 bits (data bus) output
 swap(data); // we need high part first for this
 LCD_PORT |= data & 0x0f; // just put out low part which was high bits of input
 delay_cycles (LCD_STL); // wait for command to settle
 LCD_EN = TRUE;    // begin first write cycle
 delay_cycles(LCD_PW);  // wait for slow LCD cmos logic to get it
 LCD_EN = FALSE;  // this is what clocks it into the LCD
 delay_cycles(LCD_STL); // hold time for slow LCD logic
 // do second nybble
 swap (data);     // swap nybbles back to get low in low again
 LCD_PORT &= 0xf0; // clear low part so we can or in the cmd/data bits
 LCD_PORT |= data & 0x0f; // just put out low part which was high bits of input
 delay_cycles (LCD_STL); // wait for command or data to settle
 LCD_EN = TRUE;    // begin second write cycle
 delay_cycles(LCD_PW);  // wait for slow LCD cmos logic to get it
 LCD_EN = FALSE;  // this is what clocks it into the LCD
 delay_cycles(LCD_STL); // hold time for slow LCD logic
}
///////////////////////////////////////////
void LCD_init()
{
// setup lcd module for 4 bit etc
// you cycle counters should remember this is only at init time, and in this
// case it pays to go slow enough to get a reliable initialization

 restart_WDT(); // just in case someone enabled it
 delay_ms(100); // give power supply some time to come up, this is called early
 restart_WDT(); // just in case someone enabled it 
 output_e(0); // all zeros for the moment 
 set_tris_e (0x80); // all outputs but ccp2
 
 output_f (0); //@@@ may mess with other uses of F, we only use F.7
 set_tris_f (0x7f); // for second set of LCD lines, all input but f.7
 
 delay_ms(5); // may or may not need these sprinkled liberally...
 // note, not only are these displays slow, but some are really slow...YMMV
 LCD_write(0x20,LCD_CMD); // set data length
 delay_us(100);           // let it take effect
 LCD_write(0x20,LCD_CMD);
 delay_us(100);
 LCD_write(0x80,LCD_CMD);
 delay_us(100);
 
 LCD_write(0x28,LCD_CMD); // set for 4 bits, 2 line/5x8, false means command
 delay_us(100);           // wait awhile for it to process command
 LCD_write(0x08,LCD_CMD); // turn it all off
 delay_us(100);           // wait awhile for slow peripheral
 LCD_write(0x0c,LCD_CMD); // display on now, no cursor
 delay_us(100);           // wait awhile for slow peripheral
 LCD_write(0x01,LCD_CMD); // clear display
 restart_WDT();           // just in case someone enabled it
 delay_ms(5);             // this takes a really long time sometimes
 LCD_write (0x06,LCD_CMD);// entry mode increment, shift off
 LCD_waitbusy();
}
////////////////////////////////////////////
// Write a string to a location on the LCD
// char position
void LCDOutXY(char x, char y, char* str)
{   // bits 0-3 of pos contain column, 4-7 contain row.
  // remap input position to match dumb 2 line lcd display
  // (change below for a 4 or more line display)
  if (y) // interesting little gap in display address space
     LCD_write(0xc0 + (x & 0x3f),LCD_CMD); // ouput row and column
  else 
     LCD_write(0x80 + (x & 0x3f),LCD_CMD); // if row zero
 
  LCD_waitbusy(); // let it happen
  x=0; // reuse free char for something else
  while(str[x] != 0) // stop at null terminator
  {
     LCD_write(str[x++],LCD_DATA); // actually put out a char
  }  
}
//////////////////////////////////////////////////////////////

void LCDCursorXY(char x, char y, char flags)
{ //flags == cursor on = 0x02, blink on = 0x01
  // just 2 gets non blink underline
  // just 1 gets blinking box
  // 3 gets blinking box
  // zero is cursor off
  LCD_waitbusy(); // in case this is called right after another lcd command
  if (y)
     LCD_write(0xc0 + (x & 0x3f),LCD_CMD); // ouput row and column
  else 
     LCD_write(0x80 + (x & 0x3f),LCD_CMD); // if row zero
  LCD_waitbusy(); // let it happen
  flags &= 0x3; // safety in case dumb caller
  LCD_write(0x0c + flags,LCD_CMD);
}
//////////////////////////////////////////////////////////////
void LCDClear()
{
 LCD_waitbusy();  // make sure it's safe to do a command, first
 LCD_write(0x01,LCD_CMD); // just clear screen
 LCD_waitbusy();   // this one takes a loooong time, so do the wait here.
}
//////////////////////////////////////////////////////////////

